#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/Xatom.h>

#include "x_color.h"
#include "state.h"
struct astime_state state;

#include "dof.xpm"
#include "months.xpm"
#include "digits.xpm"

/* nice idea from ascd */
typedef struct _XpmIcon {
    Pixmap pixmap;
    Pixmap mask;
    XpmAttributes attributes;
} XpmIcon;

XpmIcon dofXpm;
XpmIcon monthsXpm;
XpmIcon digitsXpm;

/* X windows related global variables */
Display * mainDisplay = 0;      /* The display we are working on */
Window Root;                    /* The root window of X11 */
Pixmap backgroundWindow;        /* Window for drawing */
Pixmap drawWindow;              /* Window for drawing */
Window mainWindow;              /* Application window */
Window iconWindow;              /* Icon window */
XGCValues mainGCV;              /* graphics context values */
GC mainGC;                      /* Graphics context */
Atom wm_delete_window;
Atom wm_protocols;

/* background, digits, months etc pixmap colors */
char bgpixmap_color[4][50];
char dof_pixmap_color[2][50];
char digits_pixmap_color[2][50];
char months_pixmap_color[2][50];

Pixel back_pix, fore_pix;

/* The center of the clock */
struct coordinates center;
/* The half-size of the watch drawable area 
 * (2 more pixels are given to the frame */
struct coordinates watch_size;

/* calculated values in pixels */
int pixelLength[3];
struct line_struct {
	int x1, y1;
	int x2, y2;
} lines[7];

/* pixels we need */
Pixel pix[3];

/* last time we updated */
time_t last_time = 0;
time_t curTime;

/* current time in nice format */
struct tm * loc_time;

/* requests for update */
int update_request = 0;

float pi=3.14159265359;

void draw_window(Window win)
{
	int i;

        XCopyArea(
                mainDisplay,
                backgroundWindow,
                win,
                mainGC,
                0,
                0,
                state.total_size.x,
                state.total_size.y,
                0,
                0
                );
	if ( state.draw_date ) {
		XCopyArea(
			mainDisplay,
			dofXpm.pixmap,
			win,
			mainGC,
			0,
			loc_time->tm_wday * 6,
			13,
			7,
			1,
			1
			);
		XCopyArea(
			mainDisplay,
			digitsXpm.pixmap,
			win,
			mainGC,
			0,
			(loc_time->tm_mday /10) * 7,
			7,
			8,
			1,
			state.total_size.y - 9
			);
		XCopyArea(
			mainDisplay,
			digitsXpm.pixmap,
			win,
			mainGC,
			0,
			(loc_time->tm_mday %10) * 7,
			7,
			8,
			8,
			state.total_size.y - 9
			);
		XCopyArea(
			mainDisplay,
			monthsXpm.pixmap,
			win,
			mainGC,
			0,
			loc_time->tm_mon * 6,
			19,
			7,
			18,
			state.total_size.y - 8
			);
		/* The year won't fit in smaller displays */
		if (state.total_size.x > 51) {
		XCopyArea(
			mainDisplay,
			digitsXpm.pixmap,
			win,
			mainGC,
			0,
			(loc_time->tm_year %100 /10) * 7,
			7,
			8,
			38,
			state.total_size.y - 9
			);
		XCopyArea(
			mainDisplay,
			digitsXpm.pixmap,
			win,
			mainGC,
			0,
			(loc_time->tm_year %10) * 7,
			7,
			8,
			45,
			state.total_size.y - 9
			);
		}
	}
#ifdef DEBUG
	mainGCV.foreground = pix[2];
        XChangeGC(
                mainDisplay,
                mainGC,
                GCForeground,
                &mainGCV
                );
	XDrawLine(
		mainDisplay,
		win,
		mainGC,
		center.x - watch_size.x,
		center.y - watch_size.y,
		center.x + watch_size.x,
		center.y - watch_size.y
		);
	XDrawLine(
		mainDisplay,
		win,
		mainGC,
		center.x + watch_size.x,
		center.y - watch_size.y,
		center.x + watch_size.x,
		center.y + watch_size.y
		);
	XDrawLine(
		mainDisplay,
		win,
		mainGC,
		center.x + watch_size.x,
		center.y + watch_size.y,
		center.x - watch_size.x,
		center.y + watch_size.y
		);
	XDrawLine(
		mainDisplay,
		win,
		mainGC,
		center.x - watch_size.x,
		center.y + watch_size.y,
		center.x - watch_size.x,
		center.y - watch_size.y
		);
#endif
	/* the pointers */
	for (i=2; i>=0; --i) {
		if ( ( i == 0 ) && ( state.draw_seconds == 0 ) ) break;
		mainGCV.foreground = pix[i];
		mainGCV.line_width = i;
		XChangeGC(
			mainDisplay,
			mainGC,
			GCForeground|GCLineWidth,
			&mainGCV
			);
		XDrawLine(
			mainDisplay,
			win,
			mainGC,
			lines[i].x1,
			lines[i].y1,
			lines[i].x2,
			lines[i].y2
			);
	}
}

/* convert the position on 0-60 scale into xy coordinates */
inline float convert2x(int tim)
{
	return cos(pi * ( ((float)(tim - 15)) * 6 ) / 180 );
}
inline float convert2y(int tim)
{
	return sin(pi * ( ((float)(tim - 15)) * 6 ) / 180 );
}

/* calculate the lines to be drawn */
void calculate_pointers()
{
	float xrate, yrate;
	loc_time = localtime(&curTime);
	/* seconds */
	xrate = convert2x(loc_time->tm_sec);
	yrate = convert2y(loc_time->tm_sec);
	lines[0].x1 = center.x - rint(5 * xrate);
	lines[0].y1 = center.y - rint(5 * yrate);
	lines[0].x2 = rint(watch_size.x * xrate) + center.x;
	lines[0].y2 = rint(watch_size.y * yrate) + center.y;
#ifdef DEBUG
	printf("%d sec = %d %d -> %d %d\n", 
		loc_time->tm_sec, lines[0].x1, lines[0].y1,
		lines[0].x2, lines[0].y2);
#endif
	/* minutes */
	xrate = convert2x(loc_time->tm_min);
	yrate = convert2y(loc_time->tm_min);
	lines[1].x1 = center.x - rint(2 * xrate);
	lines[1].y1 = center.y - rint(2 * yrate);
	lines[1].x2 = rint((watch_size.x - 1) * xrate) + center.x;
	lines[1].y2 = rint((watch_size.y - 1) * yrate) + center.y;
	/* hours */
	xrate = convert2x((loc_time->tm_hour % 12) * 5 + loc_time->tm_min/12);
	yrate = convert2y((loc_time->tm_hour % 12) * 5 + loc_time->tm_min/12);
	lines[2].x1 = center.x - rint(2 * xrate);
	lines[2].y1 = center.y - rint(2 * yrate);
	lines[2].x2 = rint((watch_size.x - 6) * xrate) + center.x;
	lines[2].y2 = rint((watch_size.y - 6) * yrate) + center.y;
}

/*
 * Calculate and redraw the lines corresponding to the
 * time.
 */
void refresh()
{
	calculate_pointers();
	draw_window(drawWindow);
	++update_request;
}

/*
 * This function clears up all X related
 * stuff and exits. It is called in case
 * of emergencies too.
 */             
void x_cleanup()
{
        if ( mainDisplay )
        {
                XCloseDisplay(mainDisplay);
        }
        exit(0);
}

/* 
 * This checks for X11 events. We distinguish the following:
 * - request to repaint the window
 * - request to quit (Close button)
 */
void CheckX11Events()
{
        XEvent Event;
        while (XPending(mainDisplay)) { 
                XNextEvent(mainDisplay, &Event);
                switch(Event.type)
                {
                case Expose:
#ifdef DEBUG
			printf("Expose event caught: (%d %d) (%d x %d)\n",
				((XExposeEvent *)&Event)->x, 
				((XExposeEvent *)&Event)->y, 
				((XExposeEvent *)&Event)->width, 
				((XExposeEvent *)&Event)->height);
#endif
                        if(Event.xexpose.count == 0) {
				++update_request;
			}
                        break;
                case ClientMessage:
                        if ((Event.xclient.message_type == wm_protocols)
                          && (Event.xclient.data.l[0] == wm_delete_window))
			{
#ifdef DEBUG
				printf("caught wm_delete_window, closing\n");
#endif
                                x_cleanup();
			}
                        break;
                }
        }
}

void redraw()
{
        XCopyArea(
                mainDisplay,
                drawWindow,
                mainWindow,
                mainGC,
                0,
                0,
		state.total_size.x,
                state.total_size.y,
                0,
                0
                );
        XCopyArea(
                mainDisplay,
                drawWindow,
                iconWindow,
                mainGC,
                0,
                0,
                state.total_size.x,
                state.total_size.y,
                0,
                0
                );
	update_request = 0;
}

void update()
{
	curTime = time(0);
	if ( (curTime - last_time) >= state.update_interval ) {
		last_time = curTime;
		refresh();
	}
	CheckX11Events();
	if ( update_request ) {
		redraw();
	}
}

/* Draw those little points every 5 secs on the background pixmap
 * that usually are marked 1,2,3,4,5,6,7 etc.
 * And those triangles that mark 3,6,9,12 o'clock */
void draw_dial()
{
	int tmp;
	struct coordinates tmp_coord;
	float xrate, yrate;

	mainGCV.foreground = fore_pix;
        XChangeGC(
                mainDisplay,
                mainGC,
                GCForeground,
                &mainGCV
                );
	for (tmp = 0; tmp < 60; tmp += 5) {
		xrate = convert2x(tmp);
		yrate = convert2y(tmp);
		tmp_coord.x = rint((watch_size.x+1) * xrate) + center.x;
		tmp_coord.y = rint((watch_size.y+1) * yrate) + center.y;
#ifdef DEBUG
		printf("dot at %d %d\n", tmp_coord.x, tmp_coord.y);
#endif
		XDrawPoint(
			mainDisplay,
			backgroundWindow,
			mainGC,
			tmp_coord.x,
			tmp_coord.y
			);
	}
	XDrawPoint(
		mainDisplay,
		backgroundWindow,
		mainGC,
		center.x-watch_size.x-1,
		center.y-1
		);
	XDrawPoint(
		mainDisplay,
		backgroundWindow,
		mainGC,
		center.x-watch_size.x-1,
		center.y+1
		);
	XDrawPoint(
		mainDisplay,
		backgroundWindow,
		mainGC,
		center.x-watch_size.x,
		center.y
		);
	XDrawPoint(
		mainDisplay,
		backgroundWindow,
		mainGC,
		center.x+watch_size.x+1,
		center.y-1
		);
	XDrawPoint(
		mainDisplay,
		backgroundWindow,
		mainGC,
		center.x+watch_size.x+1,
		center.y+1
		);
	XDrawPoint(
		mainDisplay,
		backgroundWindow,
		mainGC,
		center.x+watch_size.x,
		center.y
		);
	XDrawPoint(
		mainDisplay,
		backgroundWindow,
		mainGC,
		center.x-1,
		center.y-watch_size.y-1
		);
	XDrawPoint(
		mainDisplay,
		backgroundWindow,
		mainGC,
		center.x+1,
		center.y-watch_size.y-1
		);
	XDrawPoint(
		mainDisplay,
		backgroundWindow,
		mainGC,
		center.x,
		center.y-watch_size.y
		);
	XDrawPoint(
		mainDisplay,
		backgroundWindow,
		mainGC,
		center.x-1,
		center.y+watch_size.y+1
		);
	XDrawPoint(
		mainDisplay,
		backgroundWindow,
		mainGC,
		center.x+1,
		center.y+watch_size.y+1
		);
	XDrawPoint(
		mainDisplay,
		backgroundWindow,
		mainGC,
		center.x,
		center.y+watch_size.y
		);
}

void initialize(
			int argc, char** argv,
			char * display_name,
			char * mainGeometry,
                        int withdrawn,
                        int iconic,
                        int pushed_in)
{
	int screen;
	Status status;
	XWindowAttributes winAttr;
	XSizeHints SizeHints;
	XTextProperty title;
	char * app_name = "astime";
	XClassHint classHint;
	int gravity;
	XWMHints WmHints;
	XEvent Event;
	int color_depth;
	Pixel tmp_pix;
	int result;

	if (strlen(mainGeometry)) {
		/* Check the user-specified size */
		result = XParseGeometry(
			mainGeometry,
			&SizeHints.x,
			&SizeHints.y,
			&SizeHints.width,
			&SizeHints.height
			);
		if (result & WidthValue) {
			state.total_size.x = SizeHints.width;
		}
		if (result & HeightValue) {
			state.total_size.y = SizeHints.height;
		}
#ifdef DEBUG
		printf("User size: %d x %d, position: %d %d\n", 
			state.total_size.x, state.total_size.y,
			SizeHints.x, SizeHints.y);
#endif
	}
	if (state.draw_date) {
		center.x = state.total_size.x / 2;
		center.y = state.total_size.y / 2 - 4;
		watch_size.x = center.x - 3;
		watch_size.y = state.total_size.y / 2 - 7;
	} else {
		center.x = state.total_size.x / 2;
		center.y = state.total_size.y / 2;
		watch_size.x = center.x - 3;
		watch_size.y = center.y - 3;
	}

	mainDisplay = XOpenDisplay(display_name);
	if ( ! mainDisplay ) {
		printf("astime : grrrr... can't open display %s. Sorry ...\n",
			XDisplayName(display_name));
		exit(1);
	}
	screen = DefaultScreen(mainDisplay);
	Root = RootWindow(mainDisplay, screen);
	back_pix = GetColor(state.bgcolor, mainDisplay, Root);
	fore_pix = GetColor(state.fgcolor, mainDisplay, Root);
	color_depth = DefaultDepth(mainDisplay, screen);
#ifdef DEBUG    
        printf("astime : detected color depth %d bpp, using %d bpp\n",
                color_depth, color_depth);
#endif          

	/* The day-of-week pixmap */
	sprintf(dof_pixmap_color[0], "c c %s", state.bgcolor);
	sprintf(dof_pixmap_color[1], "l c %s", state.daycolor);
	dof[1] = dof_pixmap_color[0];
	dof[2] = dof_pixmap_color[1];
	dofXpm.attributes.valuemask |=
                (XpmReturnPixels | XpmReturnExtensions);
        status = XpmCreatePixmapFromData(
                mainDisplay,                    /* display */
                Root,                           /* window */
                dof,                     /* xpm */ 
                &dofXpm.pixmap,          /* resulting pixmap */
                &dofXpm.mask,
                &dofXpm.attributes
                );
        if(status != XpmSuccess) {
                printf("astime : (%d) not enough free color cells for day_of_week.\n", status);   
		x_cleanup();
        }       
	/* The digits pixmap */
	sprintf(digits_pixmap_color[0], "c c %s", state.bgcolor);
	sprintf(digits_pixmap_color[1], ". c %s", state.datecolor);
	digits[1] = digits_pixmap_color[0];
	digits[2] = digits_pixmap_color[1];
	digitsXpm.attributes.valuemask |=
                (XpmReturnPixels | XpmReturnExtensions);
        status = XpmCreatePixmapFromData(
                mainDisplay,                    /* display */
                Root,                           /* window */
                digits,                     /* xpm */ 
                &digitsXpm.pixmap,          /* resulting pixmap */
                &digitsXpm.mask,
                &digitsXpm.attributes
                );
        if(status != XpmSuccess) {
                printf("astime : (%d) not enough free color cells for digits.\n", status);   
		x_cleanup();
        }       
	/* The months pixmap */
	sprintf(months_pixmap_color[0], ". c %s", state.bgcolor);
	sprintf(months_pixmap_color[1], "o c %s", state.datecolor);
	months[1] = months_pixmap_color[0];
	months[2] = months_pixmap_color[1];
	monthsXpm.attributes.valuemask |=
                (XpmReturnPixels | XpmReturnExtensions);
        status = XpmCreatePixmapFromData(
                mainDisplay,                    /* display */
                Root,                           /* window */
                months,                     /* xpm */ 
                &monthsXpm.pixmap,          /* resulting pixmap */
                &monthsXpm.mask,
                &monthsXpm.attributes
                );
        if(status != XpmSuccess) {
                printf("astime : (%d) not enough free color cells for months.\n", status);   
		x_cleanup();
        }       

        SizeHints.flags= USSize|USPosition|PMinSize|PMaxSize;
        SizeHints.x = 0;
        SizeHints.y = 0;
        XWMGeometry(
                mainDisplay,
                screen,
                mainGeometry,
                NULL,
                1,
                & SizeHints,
                &SizeHints.x,
                &SizeHints.y,
                &SizeHints.width,
                &SizeHints.height,
                &gravity
                );
#ifdef DEBUG
	printf("SizeHints were: %d x %d, setting to %d x %d, position %d %d\n",
		SizeHints.width, SizeHints.height,
		state.total_size.x, state.total_size.y,
		SizeHints.x, SizeHints.y);
#endif
        SizeHints.min_width =
        SizeHints.max_width =
        SizeHints.width = state.total_size.x;
        SizeHints.min_height =
        SizeHints.max_height =
        SizeHints.height= state.total_size.y;

        backgroundWindow = XCreatePixmap(
                mainDisplay,
                Root,
                state.total_size.x,
                state.total_size.y,
		color_depth
                );
                
        drawWindow = XCreatePixmap(
                mainDisplay,            /* display */
                Root,                   /* parent */
                state.total_size.x,           /* width */
                state.total_size.y,           /* height */
		color_depth		/* color depth */
                );
                
        mainWindow = XCreateSimpleWindow(
                mainDisplay,            /* display */
                Root,                   /* parent */
                SizeHints.x,            /* x */
                SizeHints.y,            /* y */
                SizeHints.width,        /* width */
                SizeHints.height,       /* height */
                0,                      /* border_width */
                fore_pix,               /* border */
                back_pix                /* background */
                );
                
        iconWindow = XCreateSimpleWindow(
                mainDisplay,            /* display */
                Root,                   /* parent */
                SizeHints.x,            /* x */
                SizeHints.y,            /* y */
                SizeHints.width,        /* width */
                SizeHints.height,       /* height */
                0,                      /* border_width */
                fore_pix,               /* border */
                back_pix                /* background */
                );
                
        XSetWMNormalHints(mainDisplay, mainWindow, &SizeHints);
        XSetWMNormalHints(mainDisplay, iconWindow, &SizeHints);
        status = XClearWindow(mainDisplay, mainWindow);
        status = XClearWindow(mainDisplay, iconWindow);
        
        status = XGetWindowAttributes(
                mainDisplay,    /* display */
                mainWindow,     /* window */
                & winAttr       /* window_attributes_return */
                );
                
        status = XStringListToTextProperty(&app_name, 1, &title);
        XSetWMName(mainDisplay, mainWindow, &title);
        XSetWMName(mainDisplay, iconWindow, &title);
        
        classHint.res_name = "astime" ;
        classHint.res_class = "ASTIME";
        XSetClassHint(mainDisplay, mainWindow, &classHint);
        XStoreName(mainDisplay, mainWindow, "astime");
        XSetIconName(mainDisplay, mainWindow, "astime");

        status = XSelectInput(
                mainDisplay,    /* display */
                mainWindow,     /* window */
                ExposureMask    /* event_mask */
                );
                
        status = XSelectInput(
                mainDisplay,    /* display */
                iconWindow,     /* window */
                ExposureMask    /* event_mask */
                );
                
        /* Creating Graphics context */
        mainGCV.foreground = fore_pix;
        mainGCV.background = back_pix;
        mainGCV.graphics_exposures = False;
        mainGCV.line_style = LineSolid;
        mainGCV.fill_style = FillSolid;
        mainGCV.line_width = 1;
        mainGC = XCreateGC(
                mainDisplay,
                mainWindow, 
                GCForeground|GCBackground|GCLineWidth|
                        GCLineStyle|GCFillStyle,
                &mainGCV
                );
                
	/* Draw the background */
	mainGCV.foreground = back_pix;
        XChangeGC(
                mainDisplay,
                mainGC,
                GCForeground,
                &mainGCV
                );
	XFillRectangle(
		mainDisplay,
		backgroundWindow,
		mainGC,
		0,
		0,
		state.total_size.x,
		state.total_size.y
		);
	/* The shadow on the sides */
	if (pushed_in) {
		tmp_pix = DarkenColor(state.bgcolor, 1.6, mainDisplay, Root);
	} else {
		tmp_pix = LightenColor(state.bgcolor, 2.0, mainDisplay, Root);
	}
	mainGCV.foreground = tmp_pix;
        XChangeGC(
                mainDisplay,
                mainGC,
                GCForeground,
                &mainGCV
                );
	XDrawLine(
		mainDisplay,
		backgroundWindow,
		mainGC,
		0,
		0,
		state.total_size.x-1,
		0
		);
	XDrawLine(
		mainDisplay,
		backgroundWindow,
		mainGC,
		0,
		0,
		0,
		state.total_size.y-1
		);
	if (! pushed_in) {
		tmp_pix = DarkenColor(state.bgcolor, 1.6, mainDisplay, Root);
	} else {
		tmp_pix = LightenColor(state.bgcolor, 2.0, mainDisplay, Root);
	}
	mainGCV.foreground = tmp_pix;
        XChangeGC(
                mainDisplay,
                mainGC,
                GCForeground,
                &mainGCV
                );
	XDrawLine(
		mainDisplay,
		backgroundWindow,
		mainGC,
		state.total_size.x-1,
		state.total_size.y,
		state.total_size.x-1,
		0
		);
	XDrawLine(
		mainDisplay,
		backgroundWindow,
		mainGC,
		state.total_size.x,
		state.total_size.y-1,
		0,
		state.total_size.y-1
		);

	/* Finish up the background (the dial) */
	draw_dial();

        status = XSetCommand(mainDisplay, mainWindow, argv, argc);
        
        /* Set up the event for quitting the window */
        wm_delete_window = XInternAtom(
                mainDisplay, 
                "WM_DELETE_WINDOW",     /* atom_name */
                False                   /* only_if_exists */
                );
        wm_protocols = XInternAtom(
                mainDisplay,
                "WM_PROTOCOLS",         /* atom_name */
                False                   /* only_if_exists */
                );
        status = XSetWMProtocols(
                mainDisplay, 
                mainWindow,
                &wm_delete_window,
                1
                );
        status = XSetWMProtocols(
                mainDisplay, 
                iconWindow,
                &wm_delete_window,
                1
                );
                
        WmHints.flags = StateHint | IconWindowHint;
        WmHints.initial_state = 
                withdrawn ? WithdrawnState :
                        iconic ? IconicState : NormalState;
        WmHints.icon_window = iconWindow;
        if ( withdrawn ) {
                WmHints.window_group = mainWindow;
                WmHints.flags |= WindowGroupHint;
        }       
        if ( iconic || withdrawn ) {
                WmHints.icon_x = SizeHints.x;
                WmHints.icon_y = SizeHints.y;
                WmHints.flags |= IconPositionHint;
        }       
        XSetWMHints(
                mainDisplay,
                mainWindow, 
                &WmHints);
                
        /* Finally show the window */
        status = XMapWindow(
                mainDisplay,
                mainWindow
                );
                
	/* Get colors while waiting for Expose */
	pix[0] = GetColor(state.seccolor, mainDisplay, Root);
	pix[1] = GetColor(state.mincolor, mainDisplay, Root);
	pix[2] = GetColor(state.hourcolor, mainDisplay, Root);

	/* try to read the time */
	curTime = time(0);
	last_time = curTime;
	refresh();

        /* wait for the Expose event now */
        XNextEvent(mainDisplay, &Event); 
        /* We 've got Expose -> draw the parts of the window. */
        redraw(); 
        XFlush(mainDisplay);
}

